/*
 * File Name: Bird.cs
 * 
 * Date: 06/01/2006
 *
 * Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
 * Permission to use, copy, modify, and distribute this Program and its documentation,
 *  if any, for any purpose and without fee is hereby granted, provided that:
 *   (i) you not charge any fee for the Program, and the Program not be incorporated
 *       by you in any software or code for which compensation is expected or received;
 *   (ii) the copyright notice listed above appears in all copies;
 *   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
 *   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
 *          pertaining to distribution of the Program without specific, written prior permission. 
 */

using System;
using System.Drawing;

namespace Flock
{
	/// <summary>
	/// This class represents one of the birds, either green or blue.
	/// Birds try to flock with birds of the same color and avoid different color
	/// birds, predators, and obstacles.
	/// </summary>
	public class Bird
	{
		/// <summary>
		/// The range at which birds will separate from each other to avoid something.
		/// With low values flocks will be tighter, and birds will not separate to avoid
		/// obstacles and other birds. With higher values, their flock will be spread
		/// apart and a flock will be more likely split in two when faced with an obstacle.
		/// </summary>
		public static int SeparationRange;

		/// <summary>
		/// The range at which birds can detect the birds, predators, and obstacles
		/// </summary>
		public static int DetectionRange;

		/// <summary>
		/// Display the separation and detection ranges on the map.
		/// </summary>
		public static bool ShowRanges = false;

		/// <summary>
		/// Random number generator
		/// </summary>
		private static Random rand = new Random();

		/// <summary>
		/// The size of the map containing the birds.
		/// </summary>
		public static Size MapSize = new Size(500, 500);

		/// <summary>
		/// The current location of this bird.
		/// </summary>
		public Point Location { get; set; }

		/// <summary>
		/// The direction that the bird is facing in degrees.
		/// </summary>
		public int CurrentTheta { get; set; }

		/// <summary>
		/// The color of this bird.
		/// </summary>
		public Color BirdColor { get; set; }

		/// <summary>
		/// The speed of the bird. This is set by the simulator, and can be changed on the fly.
		/// </summary>
		public double CurrentSpeed { get; set; }

		/// <summary>
		/// Maximum turn radius.
		/// </summary>
		public int MaxTheta { get; set; }

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="x">Starting X coordinate of the Bird</param>
		/// <param name="y">Starting Y coordinate of the Bird</param>
		/// <param name="theta">Starting angle of the Bird - the direction the bird is facing in degrees</param>
		/// <param name="birdColor">The color of the bird</param>
		public Bird(int x, int y, int theta, Color birdColor)
		{
			Location = new Point(x, y);
			CurrentTheta = theta;
			BirdColor = birdColor;
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="birdColor">The color of the bird</param>
		public Bird(Color birdColor) : this(rand.Next(MapSize.Width), rand.Next(MapSize.Height), rand.Next(360), birdColor)
		{
		}

		/// <summary>
		/// Causes the bird to attempt to face a new direction.
		/// Based on maxTurnTheta, the bird may not be able to complete the turn.
		/// </summary>
		/// <param name="newHeading">The direction in degrees that the bird should turn toward.</param>
		public virtual void Move(int newHeading)
		{
			// determine if it is better to turn left or right for the new heading
			int left = (newHeading - CurrentTheta + 360) % 360;
			int right = (CurrentTheta - newHeading + 360) % 360;

			// after deciding which way to turn, find out if we can turn that much
			int thetaChange = 0;
			if (left < right)
			{
				// if left > than the max turn, then we can't fully adopt the new heading
				thetaChange = Math.Min(MaxTheta, left);
			}
			else
			{
				// right turns are negative degrees
				thetaChange = -Math.Min(MaxTheta, right);
			}

			// Make the turn
			CurrentTheta = (CurrentTheta + thetaChange + 360) % 360;

			// Now move currentSpeed pixels in the direction the bird now faces.
			// Note: Because values are truncated, a speed of 1 will result in no
			// movement unless the bird is moving exactly vertically or horizontally.
			var x = Location.X + (int)(CurrentSpeed * Math.Cos(CurrentTheta * Math.PI / 180)) + MapSize.Width;
			x %= MapSize.Width;
			var y = Location.Y - (int)(CurrentSpeed * Math.Sin(CurrentTheta * Math.PI / 180)) + MapSize.Height;
			y %= MapSize.Height;

			Location = new Point(x, y);
		}

		/// <summary>
		/// Get the distance in pixels between this bird and another
		/// </summary>
		/// <param name="otherBird">The other bird to measure the distance between</param>
		/// <returns>The distance to the other bird</returns>
		public int GetDistance(Bird otherBird)
		{
			return GetDistance(otherBird.Location);
		}

		/// <summary>
		/// Get the distance in pixels between this bird and a point
		/// </summary>
		/// <param name="p">The point to measure the distance between</param>
		/// <returns>The distance between this bird and the point</returns>
		public int GetDistance(Point p)
		{
			int dX = p.X - Location.X;
			int dY = p.Y - Location.Y;
			return (int)Math.Sqrt(Math.Pow(dX, 2) + Math.Pow(dY, 2));
		}

		/// <summary>
		/// Draw the bird
		/// </summary>
		/// <param name="g">The graphics object to draw the bird on.</param>
		public virtual void Draw(Graphics g)
		{
			using (var brush = new SolidBrush(BirdColor))
			{
				g.FillPie(brush, Location.X - 12, Location.Y - 12, 24, 24, 180 - CurrentTheta - 20, 40);
			}

			if (ShowRanges)
			{
				DrawRanges(g);
			}
		}

		/// <summary>
		/// Draws the range circles around a bird. In case the bird is close to one
		/// of the edges of the map, it will draw up to 4 circles for each bird so
		/// that the circles are visible on the other edge of the map.
		/// </summary>
		/// <param name="g">The graphics object to draw the ranges on.</param>
		public void DrawRanges(Graphics g)
		{
			DrawCircles(g, Location.X, Location.Y);

			int radius = Obstacle.DetectionRange;

			bool top = (Location.Y < DetectionRange);
			bool bottom = (Location.Y > MapSize.Height - DetectionRange);
			// if the obstacle radius overlaps an edge, then display a second obstacle
			// with a center point outside the range of the displayed area.
			// This way, the birds can react to it before wrapping, and it will also get drawn.
			if (Location.X < DetectionRange)
			{ // if left
				DrawCircles(g, MapSize.Width + Location.X, Location.Y);
				if (top)
				{
					DrawCircles(g, MapSize.Width + Location.X, MapSize.Height + Location.Y);
				}
				else if (bottom)
				{
					DrawCircles(g, MapSize.Width + Location.X, Location.Y - MapSize.Height);
				}
			}
			else if (Location.X > MapSize.Width - DetectionRange)
			{ // if right
				DrawCircles(g, Location.X - MapSize.Width, Location.Y);
				if (top)
				{
					DrawCircles(g, Location.X - MapSize.Width, MapSize.Height + Location.Y);
				}
				else if (bottom)
				{
					DrawCircles(g, Location.X - MapSize.Width, Location.Y - MapSize.Height);
				}
			}
			if (top)
			{
				DrawCircles(g, Location.X, MapSize.Height + Location.Y);
			}
			else if (bottom)
			{
				DrawCircles(g, Location.X, Location.Y - MapSize.Height);
			}
		}

		/// <summary>
		/// Draws the range circles around the bird.
		/// The inner circle is the point at which birds will separate to avoid the obstacle.
		/// The outer circle is the range at which the birds can detect the obstacle.
		/// </summary>
		/// <param name="g">The graphics object to draw the bird on.</param>
		/// <param name="x">The X coordinate of the bird</param>
		/// <param name="y">The Y coordinate of the bird</param>
		protected void DrawCircles(Graphics g, int x, int y)
		{
			var color = Color.FromArgb(BirdColor.R / 2, BirdColor.G / 2, BirdColor.B / 2);
			using (var pen = new Pen(color))
			{
				g.DrawEllipse(pen, x - DetectionRange, y - DetectionRange, 2 * DetectionRange, 2 * DetectionRange);
			}
			using (var pen = new Pen(BirdColor))
			{
				g.DrawEllipse(pen, x - SeparationRange, y - SeparationRange, 2 * SeparationRange, 2 * SeparationRange);
			}
		}
	}
}