/**
 * File Name: Flock.cs
 * 
 * Date: 06/01/2006
 *
 * Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
 * Permission to use, copy, modify, and distribute this Program and its documentation,
 *  if any, for any purpose and without fee is hereby granted, provided that:
 *   (i) you not charge any fee for the Program, and the Program not be incorporated
 *       by you in any software or code for which compensation is expected or received;
 *   (ii) the copyright notice listed above appears in all copies;
 *   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
 *   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
 *          pertaining to distribution of the Program without specific, written prior permission. 
 */

using System;
using System.Drawing;
using System.Collections.Generic;

namespace Flock
{
	/// <summary>
	/// This class represents the group of all birds, predators and obstacles.
	/// It controls the coordinated movement of the birds and the detection,
	/// separation logic.
	/// </summary>
	public class Flock {
		/// <summary>
		/// The list of birds in the map
		/// </summary>
		private List<Bird> birds;

		/// <summary>
		/// The range that birds will separate to avoid an object
		/// </summary>
		public static int SeparationRange;

		/// <summary>
		/// The range at which birds can detect an object
		/// </summary>
		public static int DetectionRange;

		/// <summary>
		/// The size of the map
		/// </summary>
		private static Size map = new Size(500, 500);

		/// <summary>
		/// Flock constructor.
		/// </summary>
		public Flock()
		{
			birds = new List<Bird>();
		}

		/// <summary>
		/// Add a bird to the flock.
		/// </summary>
		/// <param name="bird">The bird to add</param>
		public void AddBird(Bird bird)
		{
			birds.Add(bird);
		}

		/// <summary>
		/// Remove a bird from the flock. One bird of the given color is removed.
		/// </summary>
		/// <param name="color">Color of birds to remove</param>
		public void RemoveBird(Color color)
		{
			for (var i = 0; i < birds.Count; i++)
			{
				if (birds[i].BirdColor == color)
				{
					birds.RemoveAt(i);
					break;
				}
			}
		}

		/// <summary>
		/// Set the settings for all birds of a given color.
		/// Called either when the user changes a slider.
		/// </summary>
		/// <param name="color">The color of the bird to update</param>
		/// <param name="speed">The new speed of the bird</param>
		/// <param name="theta">The new theta of the bird</param>
		public void SetBirdParameters(Color color, int speed, int theta)
		{
			for (var i = 0; i < birds.Count; i++)
			{
				// if the color of the bird matches, then set the values
				if (birds[i].BirdColor == color)
				{
					birds[i].CurrentSpeed = speed;
					birds[i].MaxTheta = theta;
				}
			}
		}

		/// <summary>
		/// Set the hunger value for all predators
		/// </summary>
		/// <param name="hunger">The new hunger value for all predators</param>
		public void SetHunger(int hunger)
		{
			for (var i = 0; i < birds.Count; i++)
			{
				if (birds[i].GetType() == typeof(Predator))
				{
					Predator p = (Predator)birds[i];
					p.Hunger = hunger;
				}
			}
		}

		/// <summary>
		/// Set the new size of the area where 
		/// </summary>
		/// <param name="newSize">The other bird to measure the distance between</param>
		public static void SetMapSize(Size newSize)
		{
			map = newSize;
		}

		/// <summary>
		/// The display function simply draws each Bird in the vector based on 
		/// its current position and direction(theta).
		/// </summary>
		/// <param name="g">Graphics</param>
		public void Draw(Graphics g)
		{
			foreach (var bird in birds)
			{
				bird.Draw(g);
			}
		}

		/// <summary>
		/// Tells each Bird in the Vector to move in the direction of the generalHeading.
		/// </summary>
		/// <returns>
		/// A vector of birds that were removed. This is used to update the
		/// sliders for the number of birds still on the map.
		/// </returns>
		public List<Bird> Move()
		{
			bool predatorIsFull = false;
			bool ateBird = false;
			int movingBird = 0;
			int birdToEat = 0;

			List<Bird> removedBirds = new List<Bird>();

			// Loop through each bird to move.
			// Done this way, because the list can change in the loop
			while (movingBird < birds.Count) {
				predatorIsFull = false;
				ateBird = false;
            
				var bird = birds[movingBird];
				bird.Move(GeneralHeading(bird));
            
				birdToEat = 0;

				// Loop through every other bird to see if we are at the same location.
				// A predator will eat other birds.
				// Any bird will eat food.
				// A bird will only eat one item per turn.
				while (birdToEat < birds.Count) {
					var otherBird = birds[birdToEat];
                
					// Are both birds at the same location (and it isn't the same bird)
					if ((movingBird != birdToEat) &&
								(bird.Location.X > otherBird.Location.X - 2) &&
								(bird.Location.X < otherBird.Location.X + 2) &&
								(bird.Location.Y > otherBird.Location.Y - 2) &&
								(bird.Location.Y < otherBird.Location.Y + 2)) {
                    
						// If this bird is a predator and the other Bird is a normal bird
						//  or is the other bird food
						if (((bird.GetType() == typeof(Predator)) && (otherBird.GetType() == typeof(Bird))) || (otherBird.GetType() == typeof(Food)))
						{
							birds.Remove(otherBird);
							removedBirds.Add(otherBird);
							ateBird = true;

							// If this bird is a predator, reduce its hunger
							if ((bird.GetType() == typeof(Predator))) {
								Predator predator = (Predator) bird;
								predator.EatBird();

								// Predator is full, so remove it from map
								if (predator.Hunger <= 0)
								{
									birds.Remove(bird);
									removedBirds.Add(bird);
									predatorIsFull = true;
								}
							}

							break;
						}
					}
					birdToEat++;
				}
				bool ateBirdBeforeCurrent = ateBird && birdToEat < movingBird;
				if (ateBirdBeforeCurrent && predatorIsFull) {
					// removed a bird in front of us, so decrement the bird counter
					movingBird--;
				}
				else if (!ateBirdBeforeCurrent && !predatorIsFull) {
					 // can move counter, because no birds removed before us.
					 movingBird++;
				}
			}
        
			return removedBirds;
		}
    
		/// <summary>
		/// Add two points together, scaling both according to their weight
		/// </summary>
		/// <param name="p1">The first point to add</param>
		/// <param name="w1">The weight of the first point</param>
		/// <param name="p2">The second point to add</param>
		/// <param name="w2">The weight of the second point</param>
		/// <returns></returns>
		public Point SumPoints(Point p1, double w1, Point p2, double w2)
		{
			return new Point((int)(w1*p1.X + w2*p2.X), (int)(w1*p1.Y + w2*p2.Y));
		}

		/// <summary>
		/// Distance from the top left of the map to a given point
		/// </summary>
		/// <param name="p">The point to measure the distance to.</param>
		/// <returns></returns>
		public double SizeOfPoint(Point p)
		{
			return Math.Sqrt(Math.Pow(p.X, 2) + Math.Pow(p.Y, 2));
		}

		/// <summary>
		/// Normalize a point.
		/// </summary>
		/// <param name="p">The point to normalize.</param>
		/// <param name="n">The normalization value.</param>
		/// <returns></returns>
		public Point NormalizePoint(Point p, double n) {
			if (SizeOfPoint(p) == 0.0) {
				return p;
			}
			else {
				double weight = n / SizeOfPoint(p);
				return new Point((int)(p.X * weight), (int)(p.Y * weight));
			}
		}

		/// <summary>
		/// Sometimes, two birds are closer together if you go off one edge of the map
		/// and return on the other.This function will convert the "other point" into
		/// a point that closest to the point p, even if it is off the map.
		/// </summary>
		/// <param name="p">The point to measure the distance to.</param>
		/// <param name="otherPoint">The point to measure the distance from.</param>
		/// <returns></returns>
		public Point ClosestLocation(Point p, Point otherPoint) {
			int dX = Math.Abs(otherPoint.X - p.X);
			int dY = Math.Abs(otherPoint.Y - p.Y);
			int x = otherPoint.X;
			int y = otherPoint.Y;
        
			// now see if the distance between birds is closer if going off one
			// side of the map and onto the other.
			if ( Math.Abs(map.Width - otherPoint.X + p.X) < dX ) {
				dX = map.Width - otherPoint.X + p.X;
				x = otherPoint.X - map.Width;
			}
			if ( Math.Abs(map.Width - p.X + otherPoint.X) < dX ) {
				dX = map.Width - p.X + otherPoint.X;
				x = otherPoint.X + map.Width;
			}
        
			if ( Math.Abs(map.Height - otherPoint.Y + p.Y) < dY ) {
				dY = map.Height - otherPoint.Y + p.Y;
				y = otherPoint.Y - map.Height;
			}
			if ( Math.Abs(map.Height - p.Y + otherPoint.Y) < dY ) {
				dY = map.Height - p.Y + otherPoint.Y;
				y = otherPoint.Y + map.Height;
			}
        
			return new Point( x, y );
		}

		/// <summary>
		/// This function determines the direction a Bird will turn towards for this step.
		/// The bird looks at every other bird and obstacle on the map to see if it is
		/// within the detection range.Predator will move toward birds.Birds will
		/// avoid birds of a different color and all obstacles.
		/// </summary>
		/// <param name="bird"></param>
		/// <returns></returns>
		private int GeneralHeading(Bird bird)
		{
			if (bird.GetType() == typeof(Obstacle) || bird.GetType() == typeof(Food))
			{
				return 0;   // obstacles and food don't move
			}
        
			// Sum the location of all birds that are within our detection range
			Point target = new Point(0, 0);
			// Number of birds that are within the detection range
			int numBirds = 0;
        
			// Loop thorough each bird to see if it is within our detection range
			for (int i=0; i < birds.Count; i++)
			{
				Bird otherBird = birds[i];
				Point otherLocation = ClosestLocation(bird.Location, otherBird.Location);
            
				// get distance to the other Bird. Note, this distance accounts for
				// the fact that the shortest path may be through the edge of the map
				int distance = bird.GetDistance(otherLocation);
            
				if (bird != otherBird && distance > 0 && distance <= DetectionRange)
				{
					/*
					 * If the other bird is the same color, the algorithm tells the
					 * bird to align its direction with the other Bird. If the distance
					 * between them differs from DetectionRange then a weighted forces
					 * is applied to move it towards that distance. This force is
					 * stronger when the birds are very close or towards the limit of detection.
					 */
					if (bird.BirdColor == otherBird.BirdColor) // birds of same color attract
					{
						Point align = new Point((int)(100 * Math.Cos(otherBird.CurrentTheta * Math.PI/180)),
												(int)(-100 * Math.Sin(otherBird.CurrentTheta * Math.PI/180)));
						align = NormalizePoint(align, 100); // alignment weight is 100
						bool tooClose = (distance < SeparationRange);
						double weight = 200.0;
						if (tooClose)
						{
							weight *= Math.Pow(1 - (double) distance / SeparationRange, 2);
						}
						else
						{
							weight *= - Math.Pow((double)( distance - SeparationRange ) / ( DetectionRange - SeparationRange ), 2);
						}
						Point attract = SumPoints(otherLocation, -1.0, bird.Location, 1.0);
						attract = NormalizePoint(attract, weight); // weight is variable
						Point dist = SumPoints(align, 1.0, attract, 1.0);
						dist = NormalizePoint(dist, 100); // final weight is 100
						target = SumPoints(target, 1.0, dist, 1.0);
					}
					/*
					 * If this bird is a predator, and the other bird is a normal bird
					 * then there is again attraction, but the weight is fixed at 1.
					 */
					else if (bird.GetType() == typeof(Predator) && otherBird.GetType() == typeof(Bird))
					{
						Point dist = SumPoints(bird.Location, -1.0, otherLocation, 1.0);
						dist = NormalizePoint(dist, 1000);
						double weight = 1;
						target = SumPoints(target, 1.0, dist, weight); // weight is variable
					}
					/*
					 * If the other bird is food, then there is attraction, for both
					 * birds and predators.
					 */
					else if (otherBird.GetType() == typeof(Food))
					{
						Point dist = SumPoints(bird.Location, -1.0, otherLocation, 1.0);
						dist = NormalizePoint(dist, 1000);
						double weight = 1;
						target = SumPoints(target, 1.0, dist, weight); // weight is variable
					}
					/*
					 * If the birds are a different color (or if the other bird is actually
					 * an obstacle, the bird is repulsed with a force that is weighted
					 * according to a distance square rule.
					 */
					else
					{
						Point dist = SumPoints(bird.Location, 1.0, otherLocation, -1.0);
						dist = NormalizePoint(dist, 1000);
						double weight = Math.Pow((1 - (double)distance/DetectionRange), 2);
						target = SumPoints(target, 1.0, dist, weight); // weight is variable
					}
					numBirds++;
				}
			}
			 // if no birds are close enough to detect, continue moving is same direction.
			if (numBirds == 0)
			{
				return bird.CurrentTheta;
			}
			else  // average target points and add to position
			{
				target = SumPoints(bird.Location, 1.0, target, 1/(double)numBirds);
			}
        
			// Turn the target location into a direction in degrees
			int targetTheta = (int)(180/Math.PI * Math.Atan2(bird.Location.Y - target.Y, target.X - bird.Location.X));
			// Make sure the angle is 0-360
			return (targetTheta + 360) % 360; // angle for Bird to steer towards
		}
	}
}