﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Flock
{
	public partial class FlockForm : Form
	{
		/// <summary>
		/// The flock of birds
		/// </summary>
		private Flock flock;

		private Thread thread;

		private bool running = false;

		const int MAXIMUM_BIRDS = 50;
		const int MAXIMUM_SPEED = 30;
		const int MAXIMUM_HUNGER = 10;
		const int MAXIMUM_DISTANCE = 200;

		// These are the default values for the sliders, also used for a reset
		int DEFAULT_NUMBER_GREEN = 10;
		int DEFAULT_NUMBER_BLUE = 15;
		int DEFAULT_NUMBER_RED = 3;
		int DEFAULT_GREEN_THETA = 15;
		int DEFAULT_BLUE_THETA = 10;
		int DEFAULT_RED_THETA = 20;
		int DEFAULT_GREEN_SPEED = 5;
		int DEFAULT_BLUE_SPEED = 6;
		int DEFAULT_RED_SPEED = 7;
		int DEFAULT_RED_HUNGER = 3;
		int DEFAULT_OBSTACLE_SEPARATE = 30;
		int DEFAULT_OBSTACLE_DETECT = 60;

		// These are the values set by the sliders
		int numberOfGreenBirds;
		int numberOfBlueBirds;
		int numberOfRedPredators;
		int greenBirdSpeed;
		int blueBirdSpeed;
		int redPredatorSpeed;
		int greenBirdMaxTheta;
		int blueBirdMaxTheta;
		int redPredatorMaxTheta;
		int redPredatorHunger;
		int separateDistance;
		int detectDistance;

		public FlockForm()
		{
			InitializeComponent();
			InitTrackBars();
			ResetSimulation();

			// Two clicks will place two items.
			this.SetStyle(ControlStyles.StandardDoubleClick, false);

			resetButton.Click += ResetButton_Click;
			showRangesCheckBox.CheckStateChanged += ShowRangesCheckBox_CheckStateChanged;
			simulationPanel.MouseClick += SimulationPanel_MouseClick;
			FormClosing += FlockForm_FormClosing;

			running = true;
			thread = new Thread(RunSimulation);
			thread.Start();
		}

		private void FlockForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			running = false;
			thread.Abort();
		}

		public void InitTrackBars()
		{
			SetTrackBar(greenNumberTrackBar, 0, MAXIMUM_BIRDS);
			SetTrackBar(blueNumberTrackBar, 0, MAXIMUM_BIRDS);
			SetTrackBar(redNumberTrackBar, 0, MAXIMUM_BIRDS);
			SetTrackBar(greenSpeedTrackBar, 0, MAXIMUM_SPEED);
			SetTrackBar(blueSpeedTrackBar, 0, MAXIMUM_SPEED);
			SetTrackBar(redSpeedTrackBar, 0, MAXIMUM_SPEED);
			SetTrackBar(greenThetaTrackBar, 0, 180);
			SetTrackBar(blueThetaTrackBar, 0, 180);
			SetTrackBar(redThetaTrackBar, 0, 180);
			SetTrackBar(detectionTrackBar, 0, MAXIMUM_DISTANCE);
			SetTrackBar(separationTrackBar, 0, MAXIMUM_DISTANCE);
			SetTrackBar(hungerTrackBar, 0, MAXIMUM_HUNGER);
		}

		private void SetTrackBar(TrackBar tb, int min, int max)
		{
			tb.SmallChange = 1;
			tb.LargeChange = 5;
			tb.TickFrequency = max <= 20 ? 1 : (int)(max / 20);
			tb.Minimum = min;
			tb.Maximum = max;
			tb.ValueChanged += TrackBar_ValueChanged;
		}

		public void SetTrackBarValues()
		{
			greenNumberTrackBar.Value = numberOfGreenBirds;
			blueNumberTrackBar.Value = numberOfBlueBirds;
			redNumberTrackBar.Value = numberOfRedPredators;
			greenSpeedTrackBar.Value = greenBirdSpeed;
			blueSpeedTrackBar.Value = blueBirdSpeed;
			redSpeedTrackBar.Value = redPredatorSpeed;
			greenThetaTrackBar.Value = greenBirdMaxTheta;
			blueThetaTrackBar.Value = blueBirdMaxTheta;
			redThetaTrackBar.Value = redPredatorMaxTheta;
			detectionTrackBar.Value = detectDistance;
			separationTrackBar.Value = separateDistance;
			hungerTrackBar.Value = redPredatorHunger;
		}

		public void ResetSimulation()
		{
			numberOfGreenBirds = DEFAULT_NUMBER_GREEN;
			numberOfBlueBirds = DEFAULT_NUMBER_BLUE;
			numberOfRedPredators = DEFAULT_NUMBER_RED;
			greenBirdMaxTheta = DEFAULT_GREEN_THETA;
			blueBirdMaxTheta = DEFAULT_BLUE_THETA;
			redPredatorMaxTheta = DEFAULT_RED_THETA;
			greenBirdSpeed = DEFAULT_GREEN_SPEED;
			blueBirdSpeed = DEFAULT_BLUE_SPEED;
			redPredatorSpeed = DEFAULT_RED_SPEED;
			redPredatorHunger = DEFAULT_RED_HUNGER;
			separateDistance = DEFAULT_OBSTACLE_SEPARATE;
			detectDistance = DEFAULT_OBSTACLE_DETECT;

			flock = new Flock();
			simulationPanel.Flock = flock;

			SetTrackBarValues();

			Bird.SeparationRange = separateDistance;
			Bird.DetectionRange = detectDistance;

			Flock.SeparationRange = separateDistance;
			Flock.DetectionRange = detectDistance;

			for (int i = 0; i < numberOfGreenBirds; i++)
			{
				Bird bird = new Bird(Color.Green);
				bird.CurrentSpeed = greenBirdSpeed;
				bird.MaxTheta = greenBirdMaxTheta;
				flock.AddBird(bird);
			}
			for (int i = 0; i < numberOfBlueBirds; i++)
			{
				Bird bird = new Bird(Color.Blue);
				bird.CurrentSpeed = blueBirdSpeed;
				bird.MaxTheta = blueBirdMaxTheta;
				flock.AddBird(bird);
			}
			for (int i = 0; i < numberOfRedPredators; i++)
			{
				Predator bird = new Predator();
				bird.CurrentSpeed = redPredatorSpeed;
				bird.MaxTheta = redPredatorMaxTheta;
				bird.Hunger = redPredatorHunger;
				flock.AddBird(bird);
			}
		}

		private void TrackBar_ValueChanged(object sender, EventArgs e)
		{
			if (!running)
			{
				return;
			}
			if (sender == greenNumberTrackBar)
			{
				if (greenNumberTrackBar.Value < numberOfGreenBirds)
				{
					for (int i = 0; i < numberOfGreenBirds - greenNumberTrackBar.Value; i++)
					{
						flock.RemoveBird(Color.Green);
					}
				}
				else
				{
					for (int i = 0; i < greenNumberTrackBar.Value - numberOfGreenBirds; i++)
					{
						Bird b = new Bird(Color.Green);
						b.CurrentSpeed = greenBirdSpeed;
						b.MaxTheta = greenBirdMaxTheta;
						flock.AddBird(b);
					}
				}
				numberOfGreenBirds = greenNumberTrackBar.Value;
			}
			else if (sender == blueNumberTrackBar)
			{
				if (blueNumberTrackBar.Value < numberOfBlueBirds)
				{
					for (int i = 0; i < numberOfBlueBirds - blueNumberTrackBar.Value; i++)
					{
						flock.RemoveBird(Color.Blue);
					}
				}
				else
				{
					for (int i = 0; i < blueNumberTrackBar.Value - numberOfBlueBirds; i++)
					{
						Bird b = new Bird(Color.Blue);
						b.CurrentSpeed = blueBirdSpeed;
						b.MaxTheta = blueBirdMaxTheta;
						flock.AddBird(b);
					}
				}
				numberOfBlueBirds = blueNumberTrackBar.Value;
			}
			else if (sender == redNumberTrackBar)
			{
				if (redNumberTrackBar.Value < numberOfRedPredators)
				{
					for (int i = 0; i < numberOfRedPredators - redNumberTrackBar.Value; i++)
					{
						flock.RemoveBird(Color.Red);
					}
				}
				else
				{
					for (int i = 0; i < redNumberTrackBar.Value - numberOfRedPredators; i++)
					{
						Predator b = new Predator();
						b.CurrentSpeed = redPredatorSpeed;
						b.MaxTheta = redPredatorMaxTheta;
						b.Hunger = redPredatorHunger;
						flock.AddBird(b);
					}
				}
				numberOfRedPredators = redNumberTrackBar.Value;
			}
			else if (sender == greenSpeedTrackBar || sender == greenThetaTrackBar)
			{
				greenBirdSpeed = greenSpeedTrackBar.Value;
				greenBirdMaxTheta = greenThetaTrackBar.Value;
				flock.SetBirdParameters(Color.Green, greenBirdSpeed, greenBirdMaxTheta);
			}
			else if (sender == blueSpeedTrackBar || sender == blueThetaTrackBar)
			{
				blueBirdSpeed = blueSpeedTrackBar.Value;
				blueBirdMaxTheta = blueThetaTrackBar.Value;
				flock.SetBirdParameters(Color.Blue, blueBirdSpeed, blueBirdMaxTheta);
			}
			else if (sender == redSpeedTrackBar || sender == redThetaTrackBar || sender == hungerTrackBar)
			{
				redPredatorSpeed = redSpeedTrackBar.Value;
				redPredatorMaxTheta = redThetaTrackBar.Value;
				redPredatorHunger = hungerTrackBar.Value;
				flock.SetBirdParameters(Color.Red, redPredatorSpeed, redPredatorMaxTheta);
				flock.SetHunger(redPredatorHunger);
			}
			else if (sender == detectionTrackBar)
			{
				detectDistance = detectionTrackBar.Value;
				Bird.DetectionRange = detectDistance;
				Flock.DetectionRange = detectDistance;
				if (detectDistance < separateDistance)
				{
					separateDistance = detectDistance;
					Bird.SeparationRange = separateDistance;
					Flock.SeparationRange = separateDistance;
				}
			}
			else if (sender == separationTrackBar)
			{
				separateDistance = separationTrackBar.Value;
				Bird.SeparationRange = separateDistance;
				Flock.SeparationRange = separateDistance;
				if (detectDistance < separateDistance)
				{
					detectDistance = separateDistance;
					Bird.DetectionRange = detectDistance;
					Flock.DetectionRange = detectDistance;
				}
			}
		}

		private void ShowRangesCheckBox_CheckStateChanged(object sender, EventArgs e)
		{
			Bird.ShowRanges = showRangesCheckBox.Checked;
		}

		private void ResetButton_Click(object sender, EventArgs e)
		{
			ResetSimulation();
		}

		private void SimulationPanel_MouseClick(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				flock.AddBird(new Food(e.X, e.Y));
			}
			else
			{
				flock.AddBird(new Obstacle(e.X, e.Y));
			}
		}

		public void RunSimulation()
		{
			while (running)
			{
				Bird.MapSize = simulationPanel.Size;
				Flock.SetMapSize(simulationPanel.Size);

				var removedBirds = flock.Move();
				foreach (var bird in removedBirds)
				{
					if (bird.BirdColor == Color.Red)
					{
						numberOfRedPredators--;
					}
					else if (bird.BirdColor == Color.Blue)
					{
						numberOfBlueBirds--;
					}
					else if (bird.BirdColor == Color.Green)
					{
						numberOfGreenBirds--;
					}
				}

				simulationPanel.Invalidate();

				if (this.InvokeRequired)
				{
					this.Invoke((MethodInvoker)(() => SetTrackBarValues()));
				}
				else
				{
					SetTrackBarValues();
				}

				try
				{
					Thread.Sleep(20); // interval between steps
				}
				catch (Exception)
				{
				}
			}
		}
	}

	public class SimulationPanel : Panel
	{
		/// <summary>
		/// The flock of birds
		/// </summary>
		public Flock Flock { get; set; }

		public SimulationPanel()
		{
			this.SetStyle(ControlStyles.UserPaint, true);
			this.SetStyle(ControlStyles.DoubleBuffer, true);
			this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
		}

		protected override void OnPaint(PaintEventArgs e)
		{
			e.Graphics.FillRectangle(Brushes.White, 0, 0, Width, Height);
			e.Graphics.DrawRectangle(Pens.Black, 0, 0, Width, Height);
			if (Flock != null)
			{
				Flock.Draw(e.Graphics);
			}
		}
	}
}
